@@ -40,12 +40,21 @@ static void update_gconf_key_from_selection (GtkTreeSelection *selection,
 static void load_theme_names                (GtkTreeView        *tree_view,
 					     GList              *theme_list,
 					     gchar              *default_theme);
+static char *path_to_theme_id               (const char *path);
 
 
+static char *
+path_to_theme_id (const char *path)
+{
+  char *dirname;
+  char *result;
 
+  dirname = g_path_get_dirname(path);
+  result = g_path_get_basename(dirname);
+  g_free(dirname);
 
-
-
+  return result;
+}
 
 static void
 cb_dialog_response (GtkDialog *dialog, gint response_id)
@@ -56,7 +65,6 @@ cb_dialog_response (GtkDialog *dialog, gint response_id)
     gtk_widget_hide (GTK_WIDGET (dialog));
 }
 
-
 static void
 setup_tree_view (GtkTreeView *tree_view,
 		 GCallback    changed_callback,
@@ -116,10 +124,11 @@ window_theme_selection_changed (GtkTreeSelection *selection,
     }
 
   window_manager = gnome_wm_manager_get_current (gdk_display_get_default_screen (gdk_display_get_default ()));
-
-  wm_settings.flags = GNOME_WM_SETTING_THEME;
-  wm_settings.theme = window_theme_name;
-  gnome_window_manager_change_settings (window_manager, &wm_settings);
+  if (window_manager != NULL) {
+    wm_settings.flags = GNOME_WM_SETTING_THEME;
+    wm_settings.theme = window_theme_name;
+    gnome_window_manager_change_settings (window_manager, &wm_settings);
+  }
 
 }
 
@@ -151,9 +160,10 @@ load_theme_names (GtkTreeView *tree_view,
 				  GTK_POLICY_NEVER, GTK_POLICY_NEVER);
   gtk_widget_set_usize (swindow, -1, -1);
 
-  for (list = theme_list; list; list = list->next)
+  for (list = theme_list; list; list = list->next->next)
     {
       const char *name = list->data;
+      const char *id = list->next->data;
       GtkTreeIter iter;
       gboolean is_default;
 
@@ -166,7 +176,7 @@ load_theme_names (GtkTreeView *tree_view,
 
       gtk_list_store_set (GTK_LIST_STORE (model), &iter,
 			  THEME_NAME_COLUMN, name,
-			  THEME_ID_COLUMN, name,
+			  THEME_ID_COLUMN, id,
 			  THEME_FLAG_COLUMN, is_default,
 			  -1);
 
@@ -273,7 +283,7 @@ gnome_theme_details_init (void)
   widget = WID ("icon_install_button");
   g_signal_connect_swapped (G_OBJECT (widget), "clicked", G_CALLBACK (gnome_theme_installer_run), parent);
   widget = WID ("icon_manage_button");
-  g_signal_connect (G_OBJECT (widget), "clicked", G_CALLBACK (gnome_theme_manager_show_manage_themes), dialog);
+  g_signal_connect (G_OBJECT (widget), "clicked", G_CALLBACK (gnome_theme_manager_icon_show_manage_themes), dialog);
 
 
   g_signal_connect (G_OBJECT (parent), "response", G_CALLBACK (cb_dialog_response), NULL);
@@ -347,6 +357,8 @@ gnome_theme_details_reread_themes_from_disk (void)
   for (list = theme_list; list; list = list->next)
     {
       GnomeThemeInfo *info = list->data;
+      /* Use same string for ID as for name with GTK themes */
+      string_list = g_list_prepend (string_list, info->name);
       string_list = g_list_prepend (string_list, info->name);
     }
 
@@ -365,7 +377,14 @@ gnome_theme_details_reread_themes_from_disk (void)
   g_list_free (theme_list);
 
   /* Next, we do the window managers */
-  string_list = gnome_window_manager_get_theme_list (window_manager);
+  theme_list = window_manager ? gnome_window_manager_get_theme_list (window_manager) : NULL;
+  string_list = NULL;
+  for (list = theme_list; list; list = list->next)
+    {
+      /* Use same string for ID as for name with Window themes */
+      string_list = g_list_prepend (string_list, list->data);
+      string_list = g_list_prepend (string_list, list->data);
+    }  
   if (string_list == NULL)
     {
       have_window_theme = FALSE;
@@ -378,6 +397,7 @@ gnome_theme_details_reread_themes_from_disk (void)
       load_theme_names (GTK_TREE_VIEW (WID ("window_theme_treeview")), string_list, WINDOW_THEME_DEFAULT_NAME);
       g_list_free (string_list);
     }
+  g_list_free (theme_list);
 
   /* Third, we do the icon theme */
   theme_list = gnome_theme_icon_info_find_all ();
@@ -386,6 +406,7 @@ gnome_theme_details_reread_themes_from_disk (void)
   for (list = theme_list; list; list = list->next)
     {
       GnomeThemeIconInfo *info = list->data;
+      string_list = g_list_prepend (string_list, path_to_theme_id(info->path));
       string_list = g_list_prepend (string_list, info->name);
     }
 
@@ -499,8 +520,10 @@ gnome_theme_details_update_from_gconf (void)
 
   tree_view = WID ("window_theme_treeview");
   wm_settings.flags = GNOME_WM_SETTING_THEME;
-  gnome_window_manager_get_settings (window_manager, &wm_settings);
-  update_list_something (tree_view, wm_settings.theme);
+  if (window_manager) {
+    gnome_window_manager_get_settings (window_manager, &wm_settings);
+    update_list_something (tree_view, wm_settings.theme);
+  }
 
   tree_view = WID ("icon_theme_treeview");
   theme = gconf_client_get_string (client, ICON_THEME_KEY, NULL);
