@@ -54,16 +54,18 @@ gint n_drop_types = sizeof (drop_types) / sizeof (GtkTargetEntry);
 
 static gboolean setting_model = FALSE;
 static gboolean idle_running = FALSE;
+static gboolean initial_meta_theme_set = FALSE;
 static GdkPixbuf *default_image = NULL;
 
 static GnomeThemeMetaInfo custom_meta_theme_info = {};
+static GnomeThemeMetaInfo initial_meta_theme_info = {};
 
 /* Function Prototypes */
 static void      idle_async_func                 (GdkPixbuf          *pixbuf,
 						  gpointer            data);
 static void      list_data_free                  (gpointer            data);
 static gboolean  load_theme_in_idle              (gpointer            data);
-static void      add_pixbuf_idle                 (GtkTreeModel       *model);
+static void      add_pixbuf_idle                 (void);
 static void      load_meta_themes                (GtkTreeView        *tree_view,
 						  GList              *meta_theme_list,
 						  char               *default_theme);
@@ -84,6 +86,10 @@ static void      icon_key_changed                (GConfClient        *client,
 						  guint               cnxn_id,
 						  GConfEntry         *entry,
 						  gpointer            user_data);
+static void      font_key_changed                (GConfClient        *client,
+						  guint               cnxn_id,
+						  GConfEntry         *entry,
+						  gpointer            user_data);
 static void      theme_changed_func              (gpointer            uri,
 						  gpointer            user_data);
 static void      cb_dialog_response              (GtkDialog          *dialog,
@@ -127,7 +133,7 @@ idle_async_func (GdkPixbuf *pixbuf,
     }
 
   idle_running = FALSE;
-  add_pixbuf_idle (model);
+  add_pixbuf_idle ();
 }
 
 static void
@@ -144,10 +150,16 @@ list_data_free (gpointer data)
 static gboolean
 load_theme_in_idle (gpointer data)
 {
-  GtkTreeModel *model = data;
+  GladeXML *dialog;
+  GtkWidget *tree_view;
+  GtkTreeModel *model;
   GtkTreeIter iter;
   gboolean valid;
 
+  dialog = gnome_theme_manager_get_theme_dialog ();
+  tree_view = WID ("meta_theme_treeview");
+  model = gtk_tree_view_get_model (GTK_TREE_VIEW (tree_view));
+
   for (valid = gtk_tree_model_get_iter_first (model, &iter);
        valid;
        valid = gtk_tree_model_iter_next (model, &iter))
@@ -169,7 +181,6 @@ load_theme_in_idle (gpointer data)
 	  if (theme_id == NULL)
 	    {
 	      meta_theme_info = &custom_meta_theme_info;
-	      continue;
 	    }
 	  else
 	    {
@@ -195,6 +206,7 @@ load_theme_in_idle (gpointer data)
    * dialog if it hasn't been done yet.  If it has, then this call is harmless.
    */
   gnome_theme_details_init ();
+  idle_running = FALSE;
 
   return FALSE;
 }
@@ -203,20 +215,42 @@ load_theme_in_idle (gpointer data)
  * during the pixbuf generation.
  */
 static void
-add_pixbuf_idle (GtkTreeModel *model)
+add_pixbuf_idle (void)
 {
   if (idle_running)
     return;
 
   idle_running = TRUE;
-  g_object_ref (model);
   g_idle_add_full (G_PRIORITY_LOW,
 		   load_theme_in_idle,
-		   model,
-		   g_object_unref);
+		   NULL, NULL);
 }
 
-/* Loads up a list of GnomeMetaThemeInfo.
+static gint
+sort_meta_theme_list_func (gconstpointer  a,
+			   gconstpointer  b)
+{
+  const GnomeThemeMetaInfo *a_meta_theme_info = a;
+  const GnomeThemeMetaInfo *b_meta_theme_info = b;
+  guint a_flag = 0;
+  guint b_flag = 0;
+
+  g_assert (a_meta_theme_info->name);
+  g_assert (b_meta_theme_info->name);
+
+  if (! strcmp (META_THEME_DEFAULT_NAME, a_meta_theme_info->name))
+    a_flag |= THEME_FLAG_DEFAULT;
+  if (! strcmp (META_THEME_DEFAULT_NAME, b_meta_theme_info->name))
+    b_flag |= THEME_FLAG_DEFAULT;
+
+  return gnome_theme_manager_sort_func (a_meta_theme_info->readable_name,
+					b_meta_theme_info->readable_name,
+					a_flag,
+					b_flag);
+}
+
+
+/* Loads up a list of GnomeThemeMetaInfo.
  */
 static void
 load_meta_themes (GtkTreeView *tree_view,
@@ -226,60 +260,208 @@ load_meta_themes (GtkTreeView *tree_view,
   GList *list;
   GtkTreeModel *model;
   GtkWidget *swindow;
+  GtkTreeIter iter;
+  gchar *name;
+  gboolean valid;
+  guint flag;
   gint i = 0;
+  GConfClient *client;
+  gchar *current_gtk_theme;
+  gchar *current_window_theme;
+  gchar *current_icon_theme;
+  GnomeWindowManager *window_manager;
+  GnomeWMSettings wm_settings;
+  static gboolean first_time = TRUE;
 
   swindow = GTK_WIDGET (tree_view)->parent;
   model = gtk_tree_view_get_model (tree_view);
   g_assert (model);
 
   setting_model = TRUE;
-  gtk_list_store_clear (GTK_LIST_STORE (model));
+
+  client = gconf_client_get_default ();
+
+  current_gtk_theme = gconf_client_get_string (client, GTK_THEME_KEY, NULL);
+  current_icon_theme = gconf_client_get_string (client, ICON_THEME_KEY, NULL);
+  window_manager = gnome_wm_manager_get_current (gdk_display_get_default_screen (gdk_display_get_default ()));
+  wm_settings.flags = GNOME_WM_SETTING_THEME;
+  gnome_window_manager_get_settings (window_manager, &wm_settings);
+  current_window_theme = g_strdup (wm_settings.theme);
+
+  /* FIXME: What do we really do when there is no theme? */
+  if (current_icon_theme == NULL)
+    current_icon_theme = g_strdup ("Default");
+  if (current_gtk_theme == NULL)
+    current_gtk_theme = g_strdup ("Default");
+
+
+  /* handle first time */
+  if (first_time)
+    {
+      initial_meta_theme_info.gtk_theme_name = g_strdup (current_gtk_theme);
+      initial_meta_theme_info.icon_theme_name = g_strdup (current_icon_theme);
+      initial_meta_theme_info.metacity_theme_name = g_strdup (current_window_theme);
+    }
 
   gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (swindow),
 				  GTK_POLICY_NEVER, GTK_POLICY_NEVER);
   gtk_widget_set_usize (swindow, -1, -1);
 
-  for (list = meta_theme_list; list; list = list->next)
+  /* Sort meta_theme_list to be in the same order of the current data.  This way
+   * we can walk through them together. */
+  meta_theme_list = g_list_sort (meta_theme_list, sort_meta_theme_list_func);
+
+  list = meta_theme_list;
+  valid = gtk_tree_model_get_iter_first (model, &iter);
+
+  while (valid || list != NULL)
     {
-      GnomeThemeMetaInfo *meta_theme_info = list->data;
+      GnomeThemeMetaInfo *list_meta_theme_info = NULL;
+      GnomeThemeMetaInfo *model_meta_theme_info = NULL;
       gchar *blurb;
-      GtkTreeIter iter;
-      gboolean is_default;
+      gboolean list_is_default = FALSE;
       GdkPixbuf *pixbuf;
-      gtk_list_store_prepend (GTK_LIST_STORE (model), &iter);
 
-      if (strcmp (default_theme, meta_theme_info->name) == 0)
-	is_default = TRUE;
-      else
-	is_default = FALSE;
+      /* Check info on the list */
+      if (list)
+	{
+	  list_meta_theme_info = list->data;
+	  if (strcmp (default_theme, list_meta_theme_info->name) == 0)
+	    list_is_default = TRUE;
+	  else
+	    list_is_default = FALSE;
+	}
 
-      blurb = g_strdup_printf ("<span size=\"larger\" weight=\"bold\">%s</span>\n%s",
-			       meta_theme_info->readable_name, meta_theme_info->comment);
-      if (i <= MAX_ELEMENTS_BEFORE_SCROLLING)
-	pixbuf = generate_theme_thumbnail (meta_theme_info);
-      else
-	pixbuf = default_image;
+      if (valid)
+	{
+	  gtk_tree_model_get (model, &iter,
+			      META_THEME_ID_COLUMN, &name,
+			      META_THEME_FLAG_COLUMN, &flag,
+			      -1);
+	  if (flag & THEME_FLAG_CUSTOM)
+	    model_meta_theme_info = &custom_meta_theme_info;
+	  else if (flag & THEME_FLAG_INITIAL)
+	    model_meta_theme_info = &initial_meta_theme_info;
+	  else
+	    model_meta_theme_info = gnome_theme_meta_info_find (name);
+	  g_free (name);
+	}
+      
+      /* start comparing values */
+      if (list && valid)
+	{
+	  gint compare_val;
 
-      gtk_list_store_set (GTK_LIST_STORE (model), &iter,
-			  META_THEME_PIXBUF_COLUMN, pixbuf,
-			  META_THEME_NAME_COLUMN, blurb,
-			  META_THEME_ID_COLUMN, meta_theme_info->name,
-			  META_THEME_FLAG_COLUMN, is_default ? THEME_FLAG_DEFAULT : 0,
-			  -1);
-      g_free (blurb);
+	  if (flag & THEME_FLAG_CUSTOM || flag & THEME_FLAG_INITIAL)
+	    {
+	      /* We can always skip the custom row, as it's never in the list */
+	      valid = gtk_tree_model_iter_next (model, &iter);
+	      i++;
+	      goto end_of_loop;
+	    }
 
+	  compare_val = gnome_theme_manager_sort_func (model_meta_theme_info->readable_name, list_meta_theme_info->readable_name,
+						       flag, list_is_default ? THEME_FLAG_DEFAULT : 0);
+
+	  if (compare_val < 0)
+	    {
+	      GtkTreeIter iter_to_remove;
+
+	      /* This item is no longer in the list
+	       */
+	      iter_to_remove = iter;
+	      valid = gtk_tree_model_iter_next (model, &iter);
+	      gtk_list_store_remove (GTK_LIST_STORE (model), &iter_to_remove);
+	    }
+	  else if (compare_val == 0)
+	    {
+	      /* We reset the blurb in case it has changed */
+	      blurb = g_strdup_printf ("<span size=\"larger\" weight=\"bold\">%s</span>\n%s",
+				       list_meta_theme_info->readable_name, list_meta_theme_info->comment);
+	      pixbuf = default_image;
+
+	      gtk_list_store_set (GTK_LIST_STORE (model), &iter,
+				  META_THEME_PIXBUF_COLUMN, pixbuf,
+				  META_THEME_NAME_COLUMN, blurb,
+				  META_THEME_ID_COLUMN, list_meta_theme_info->name,
+				  META_THEME_FLAG_COLUMN, flag,
+				  -1);
+	      g_free (blurb);
+
+	      list = list->next;
+	      valid = gtk_tree_model_iter_next (model, &iter);
+	      i++;
+	    }
+	  else
+	    {
+	      /* we insert a new item */
+	      GtkTreeIter iter_to_prepend;
+
+	      gtk_list_store_insert_before (GTK_LIST_STORE (model), &iter_to_prepend, &iter);
+	      blurb = g_strdup_printf ("<span size=\"larger\" weight=\"bold\">%s</span>\n%s",
+				       list_meta_theme_info->readable_name, list_meta_theme_info->comment);
+	      pixbuf = default_image;
+
+	      gtk_list_store_set (GTK_LIST_STORE (model), &iter,
+				  META_THEME_PIXBUF_COLUMN, pixbuf,
+				  META_THEME_NAME_COLUMN, blurb,
+				  META_THEME_ID_COLUMN, list_meta_theme_info->name,
+				  META_THEME_FLAG_COLUMN, list_is_default ? THEME_FLAG_DEFAULT : 0,
+				  -1);
+	      g_free (blurb);
+
+	      i++;
+	      list = list->next;
+	    }
+	  goto end_of_loop;
+	}
+      else if (list)
+	{
+	  list_meta_theme_info = list->data;
+	  gtk_list_store_append (GTK_LIST_STORE (model), &iter);
+
+	  blurb = g_strdup_printf ("<span size=\"larger\" weight=\"bold\">%s</span>\n%s",
+				   list_meta_theme_info->readable_name, list_meta_theme_info->comment);
+	  if (i <= MAX_ELEMENTS_BEFORE_SCROLLING)
+	    pixbuf = generate_theme_thumbnail (list_meta_theme_info, FALSE);
+	  else
+	    pixbuf = default_image;
+
+	  gtk_list_store_set (GTK_LIST_STORE (model), &iter,
+			      META_THEME_PIXBUF_COLUMN, pixbuf,
+			      META_THEME_NAME_COLUMN, blurb,
+			      META_THEME_ID_COLUMN, list_meta_theme_info->name,
+			      META_THEME_FLAG_COLUMN, list_is_default ? THEME_FLAG_DEFAULT : 0,
+			      -1);
+	  g_free (blurb);
+
+	  list = list->next;
+	  i++;
+	}
+      else if (valid)
+	{
+	  /* It's a dead item. */
+	  GtkTreeIter iter_to_remove;
+
+	  iter_to_remove = iter;
+	  valid = gtk_tree_model_iter_next (model, &iter);
+	  gtk_list_store_remove (GTK_LIST_STORE (model), &iter_to_remove);
+	}
+
+    end_of_loop:
       if (i == MAX_ELEMENTS_BEFORE_SCROLLING)
 	{
-	  GtkRequisition rectangle;
-	  gtk_widget_size_request (GTK_WIDGET (tree_view), &rectangle);
-	  gtk_widget_set_usize (swindow, -1, rectangle.height);
 	  gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (swindow),
 					  GTK_POLICY_NEVER, GTK_POLICY_AUTOMATIC);
 	}
-      i++;
     }
 
-  add_pixbuf_idle (model);
+  add_pixbuf_idle ();
+
+  g_free (current_gtk_theme);
+  g_free (current_icon_theme);
+  g_free (current_window_theme);
+  first_time = FALSE;
   setting_model = FALSE;
 }
 
@@ -337,7 +519,8 @@ meta_theme_set (GnomeThemeMetaInfo *meta_theme_info)
   /* Set the wm key */
   wm_settings.flags = GNOME_WM_SETTING_THEME;
   wm_settings.theme = meta_theme_info->metacity_theme_name;
-  gnome_window_manager_change_settings (window_manager, &wm_settings);
+  if (window_manager)
+    gnome_window_manager_change_settings (window_manager, &wm_settings);
 
   /* set the icon theme */
   old_key = gconf_client_get_string (client, ICON_THEME_KEY, NULL);
@@ -395,46 +578,10 @@ static void
 update_themes_from_disk (GladeXML *dialog)
 {
   GList *theme_list;
-  GList *string_list;
-  GList *list;
-  GConfClient *client;
-  gchar *current_gtk_theme;
-  gchar *current_window_theme;
-  gchar *current_icon_theme;
-  GnomeWindowManager *window_manager;
-  GnomeWMSettings wm_settings;
-  GtkWidget *notebook;
-
   gboolean have_meta_theme;
 
-  client = gconf_client_get_default ();
-
-  current_gtk_theme = gconf_client_get_string (client, GTK_THEME_KEY, NULL);
-  current_icon_theme = gconf_client_get_string (client, ICON_THEME_KEY, NULL);
-  window_manager = gnome_wm_manager_get_current (gdk_display_get_default_screen (gdk_display_get_default ()));
-  wm_settings.flags = GNOME_WM_SETTING_THEME;
-  gnome_window_manager_get_settings (window_manager, &wm_settings);
-  current_window_theme = g_strdup (wm_settings.theme);
-
-  /* FIXME: What do we really do when there is no theme?  Ask Havoc here. */
-  /* BROKEN BROKEN BROKEN */
-  if (current_icon_theme == NULL)
-    current_icon_theme = g_strdup ("Default");
-  if (current_gtk_theme == NULL)
-    current_gtk_theme = g_strdup ("Default");
-
-  notebook = WID ("theme_notebook");
-
   theme_list = gnome_theme_meta_info_find_all ();
-  string_list = NULL;
-  for (list = theme_list; list; list = list->next)
-    {
-      GnomeThemeMetaInfo *info = list->data;
-
-      string_list = g_list_prepend (string_list, info);
-    }
-
-  if (string_list == NULL)
+  if (theme_list == NULL)
     {
       have_meta_theme = FALSE;
     }
@@ -442,14 +589,10 @@ update_themes_from_disk (GladeXML *dialog)
     {
       have_meta_theme = TRUE;
       gtk_widget_show (WID ("meta_theme_hbox"));
-      load_meta_themes (GTK_TREE_VIEW (WID ("meta_theme_treeview")), string_list, META_THEME_DEFAULT_NAME);
-      g_list_free (string_list);
+      load_meta_themes (GTK_TREE_VIEW (WID ("meta_theme_treeview")), theme_list, META_THEME_DEFAULT_NAME);
     }
   g_list_free (theme_list);
 
-  g_free (current_gtk_theme);
-  g_free (current_icon_theme);
-
   update_settings_from_gconf ();
 
   if (! have_meta_theme)
@@ -479,14 +622,20 @@ add_custom_row_to_meta_theme (const gchar  *current_gtk_theme,
   GtkTreeIter iter;
   gboolean valid;
   gchar *blurb;
+  GdkPixbuf *pixbuf;
 
   dialog = gnome_theme_manager_get_theme_dialog ();
   tree_view = WID ("meta_theme_treeview");
   model = gtk_tree_view_get_model (GTK_TREE_VIEW (tree_view));
-  
+
+  g_free (custom_meta_theme_info.gtk_theme_name);
   custom_meta_theme_info.gtk_theme_name = g_strdup (current_gtk_theme);
+  g_free (custom_meta_theme_info.metacity_theme_name);
   custom_meta_theme_info.metacity_theme_name = g_strdup (current_window_theme);
+  g_free (custom_meta_theme_info.icon_theme_name);
   custom_meta_theme_info.icon_theme_name = g_strdup (current_icon_theme);
+  g_free (custom_meta_theme_info.name);
+  custom_meta_theme_info.name = g_strdup ("__Custom Theme__");
 
   for (valid = gtk_tree_model_get_iter_first (model, &iter);
        valid;
@@ -497,7 +646,7 @@ add_custom_row_to_meta_theme (const gchar  *current_gtk_theme,
       gtk_tree_model_get (model, &iter,
 			  META_THEME_FLAG_COLUMN, &theme_flags,
 			  -1);
-      if (theme_flags & THEME_FLAG_CUSTOM)
+      if (theme_flags & THEME_FLAG_CUSTOM || theme_flags & THEME_FLAG_INITIAL)
 	break;
 
     }
@@ -512,11 +661,19 @@ add_custom_row_to_meta_theme (const gchar  *current_gtk_theme,
   blurb = g_strdup_printf ("<span size=\"larger\" weight=\"bold\">%s</span>\n%s",
 			   _("Custom theme"), _("You can save this theme by pressing the Save Theme button."));
 
+  /* Invalidate the cache because the custom theme has potentially changed */
+  /* Commented out because it does odd things */
+  /*theme_thumbnail_invalidate_cache (&custom_meta_theme_info);*/
+
+  pixbuf = generate_theme_thumbnail (&custom_meta_theme_info, TRUE);
+  if (pixbuf == NULL)
+    pixbuf = default_image;
   gtk_list_store_set (GTK_LIST_STORE (model), &iter,
+		      META_THEME_PIXBUF_COLUMN, pixbuf,
 		      META_THEME_NAME_COLUMN, blurb,
 		      META_THEME_FLAG_COLUMN, THEME_FLAG_CUSTOM,
-		      META_THEME_PIXBUF_COLUMN, default_image,
 		      -1);
+
   gtk_widget_set_sensitive (WID ("meta_theme_save_button"), TRUE);
   path = gtk_tree_model_get_path (model, &iter);
   gtk_tree_view_set_cursor (GTK_TREE_VIEW (tree_view), path, NULL, FALSE);
@@ -526,14 +683,78 @@ add_custom_row_to_meta_theme (const gchar  *current_gtk_theme,
 
 
 static void
-remove_custom_row_from_meta_theme (GtkTreeModel *model)
+add_initial_row_to_meta_theme (void)
 {
   GladeXML *dialog;
+  GtkWidget *tree_view;
+  GtkTreeModel *model;
+  GtkTreePath *path;
+  GtkTreeIter iter;
+  gboolean valid;
+  gchar *blurb;
+  GdkPixbuf *pixbuf;
+
+  dialog = gnome_theme_manager_get_theme_dialog ();
+  tree_view = WID ("meta_theme_treeview");
+  model = gtk_tree_view_get_model (GTK_TREE_VIEW (tree_view));
+
+  for (valid = gtk_tree_model_get_iter_first (model, &iter);
+       valid;
+       valid = gtk_tree_model_iter_next (model, &iter))
+    {
+      guint theme_flags = 0;
+
+      gtk_tree_model_get (model, &iter,
+			  META_THEME_FLAG_COLUMN, &theme_flags,
+			  -1);
+      if (theme_flags & THEME_FLAG_CUSTOM || theme_flags & THEME_FLAG_INITIAL)
+	break;
+    }
+
+  /* if we found a custom row and broke out of the list above, valid will be
+   * TRUE.  If we didn't, we need to add a new iter.
+   */
+  if (!valid)
+    gtk_list_store_prepend (GTK_LIST_STORE (model), &iter);
+
+  /* set the values of the Custom theme. */
+  blurb = g_strdup_printf ("<span size=\"larger\" weight=\"bold\">%s</span>\n%s",
+			   _("Custom theme"), _("You can save this theme by pressing the Save Theme button."));
+
+  /* Invalidate the cache because the custom theme has potentially changed */
+  /* Commented out because it does odd things */
+  /*theme_thumbnail_invalidate_cache (&custom_meta_theme_info);*/
+
+  pixbuf = generate_theme_thumbnail (&initial_meta_theme_info, TRUE);
+  if (pixbuf == NULL)
+    pixbuf = default_image;
+
+  gtk_list_store_set (GTK_LIST_STORE (model), &iter,
+		      META_THEME_PIXBUF_COLUMN, pixbuf,
+		      META_THEME_NAME_COLUMN, blurb,
+		      META_THEME_FLAG_COLUMN, THEME_FLAG_INITIAL,
+		      -1);
+
+  gtk_widget_set_sensitive (WID ("meta_theme_save_button"), TRUE);
+  path = gtk_tree_model_get_path (model, &iter);
+  gtk_tree_path_free (path);
+  g_free (blurb);
+}
+
+
+static void
+remove_custom_row_from_meta_theme (void)
+{
+  GladeXML *dialog;
+  GtkWidget *tree_view;
+  GtkTreeModel *model;
   GtkTreeIter iter;
   GtkTreeIter next_iter;
   gboolean valid;
 
   dialog = gnome_theme_manager_get_theme_dialog ();
+  tree_view = WID ("meta_theme_treeview");
+  model = gtk_tree_view_get_model (GTK_TREE_VIEW (tree_view));
 
   valid = gtk_tree_model_get_iter_first (model, &iter);
   while (valid)
@@ -547,7 +768,7 @@ remove_custom_row_from_meta_theme (GtkTreeModel *model)
 			  META_THEME_FLAG_COLUMN, &theme_flags,
 			  -1);
 
-      if (theme_flags & THEME_FLAG_CUSTOM)
+      if (theme_flags & THEME_FLAG_CUSTOM || theme_flags & THEME_FLAG_INITIAL)
 	{
 	  gtk_list_store_remove (GTK_LIST_STORE (model), &iter);
 	}
@@ -556,12 +777,14 @@ remove_custom_row_from_meta_theme (GtkTreeModel *model)
   g_free (custom_meta_theme_info.gtk_theme_name);
   g_free (custom_meta_theme_info.metacity_theme_name);
   g_free (custom_meta_theme_info.icon_theme_name);
+  g_free (custom_meta_theme_info.name);
 
   gtk_widget_set_sensitive (WID ("meta_theme_save_button"), FALSE);
 
   custom_meta_theme_info.gtk_theme_name = NULL;
   custom_meta_theme_info.metacity_theme_name = NULL;
   custom_meta_theme_info.icon_theme_name = NULL;
+  custom_meta_theme_info.name = NULL;
 }
 
 
@@ -582,6 +805,7 @@ update_settings_from_gconf (void)
   GladeXML *dialog;
   gboolean valid;
   gboolean custom_theme_found;
+  static gboolean first_time_run = TRUE;
 
   client = gconf_client_get_default ();
 
@@ -590,8 +814,12 @@ update_settings_from_gconf (void)
   current_icon_theme = gconf_client_get_string (client, ICON_THEME_KEY, NULL);
   window_manager = gnome_wm_manager_get_current (gdk_display_get_default_screen (gdk_display_get_default ()));
   wm_settings.flags = GNOME_WM_SETTING_THEME;
-  gnome_window_manager_get_settings (window_manager, &wm_settings);
-  current_window_theme = g_strdup (wm_settings.theme);
+  if (window_manager) {
+    gnome_window_manager_get_settings (window_manager, &wm_settings);
+    current_window_theme = g_strdup (wm_settings.theme);
+  } else
+    current_window_theme = g_strdup ("");
+
   custom_theme_found = TRUE;
 
   /* Walk the tree looking for the current one. */
@@ -617,6 +845,10 @@ update_settings_from_gconf (void)
 	{
 	  meta_theme_info = &custom_meta_theme_info;
 	}
+      else if (row_theme_flags & THEME_FLAG_INITIAL)
+	{
+	  meta_theme_info = &initial_meta_theme_info;
+	}
       else if (row_theme_id)
 	{
 	  meta_theme_info = gnome_theme_meta_info_find (row_theme_id);
@@ -630,8 +862,8 @@ update_settings_from_gconf (void)
       if (meta_theme_info == &(custom_meta_theme_info))
 	custom_theme_found = TRUE;
       else if (! strcmp (current_gtk_theme, meta_theme_info->gtk_theme_name) &&
-	  ! strcmp (current_window_theme, meta_theme_info->metacity_theme_name) &&
-	  ! strcmp (current_icon_theme, meta_theme_info->icon_theme_name))
+	       ! strcmp (current_window_theme, meta_theme_info->metacity_theme_name) &&
+	       ! strcmp (current_icon_theme, meta_theme_info->icon_theme_name))
 	{
 	  GtkTreePath *path;
 
@@ -644,15 +876,33 @@ update_settings_from_gconf (void)
 	}
     }
 
-  if (custom_theme_found)
-    add_custom_row_to_meta_theme (current_gtk_theme, current_window_theme, current_icon_theme);
-  else
-    remove_custom_row_from_meta_theme (model);
+  if (custom_theme_found && first_time_run)
+    {
+      initial_meta_theme_set = TRUE;
+      initial_meta_theme_info.name = g_strdup ("__Initial Theme__");
+      initial_meta_theme_info.gtk_theme_name = g_strdup (current_gtk_theme);
+      initial_meta_theme_info.metacity_theme_name = g_strdup (current_window_theme);
+      initial_meta_theme_info.icon_theme_name = g_strdup (current_icon_theme);
+    }
+  first_time_run = FALSE;
+  
 
+  if (custom_theme_found)
+    {
+      add_custom_row_to_meta_theme (current_gtk_theme, current_window_theme, current_icon_theme);
+    }
+  else if (initial_meta_theme_set)
+    {
+      add_initial_row_to_meta_theme ();
+    }      
+  else
+    {
+      remove_custom_row_from_meta_theme ();
+    }
   g_free (current_gtk_theme);
   g_free (current_window_theme);
   g_free (current_icon_theme);
-  add_pixbuf_idle (model);
+  add_pixbuf_idle ();
 }
 
 static void
@@ -677,6 +927,62 @@ window_settings_changed (GnomeWindowManager *window_manager,
   gnome_theme_details_update_from_gconf ();
 }
 
+static void
+update_font_button_state (GladeXML *dialog)
+{
+  GConfClient *client = gconf_client_get_default ();
+  GtkTreeSelection *selection;
+  GtkTreeModel *model;
+  GtkTreeIter iter;
+  
+  selection = gtk_tree_view_get_selection (GTK_TREE_VIEW (WID ("meta_theme_treeview")));
+  
+  if (gtk_tree_selection_get_selected (selection, &model, &iter))
+    {
+      GnomeThemeMetaInfo *meta_theme_info;
+      char *meta_theme_name, *str;
+      
+      gtk_tree_model_get (model, &iter,
+			  META_THEME_ID_COLUMN, &meta_theme_name,
+			  -1);
+      if (!meta_theme_name)
+	return;
+      
+      meta_theme_info = gnome_theme_meta_info_find (meta_theme_name);
+
+      g_assert (meta_theme_info);
+      g_free (meta_theme_name);
+
+      str = gconf_client_get_string (client, FONT_KEY, NULL);
+      
+      if (meta_theme_info->application_font != NULL && str != NULL &&
+	  strcmp (meta_theme_info->application_font, str) == 0)
+	{
+	  gtk_widget_set_sensitive (WID ("meta_theme_font1_button"), FALSE);
+	  gtk_widget_set_sensitive (WID ("meta_theme_font2_button"), FALSE);
+	}
+      else
+	{
+	  gtk_widget_set_sensitive (WID ("meta_theme_font1_button"), TRUE);
+	  gtk_widget_set_sensitive (WID ("meta_theme_font2_button"), TRUE);
+	}
+
+      g_free (str);
+    }  
+  
+}
+
+static void
+font_key_changed (GConfClient        *client,
+		  guint               cnxn_id,
+		  GConfEntry         *entry,
+		  gpointer            user_data)
+{
+  GladeXML *dialog = user_data;
+
+  update_font_button_state (dialog);
+}
+
 static void
 icon_key_changed (GConfClient *client,
 		  guint        cnxn_id,
@@ -699,8 +1005,7 @@ theme_changed_func (gpointer uri,
 
   dialog = gnome_theme_manager_get_theme_dialog ();
 
-  g_print ("theme_changed_func:\n");
-  update_themes_from_disk ((GladeXML *)user_data);
+  update_themes_from_disk (dialog);
   gnome_theme_details_reread_themes_from_disk ();
   gtk_widget_grab_focus (WID ("meta_theme_treeview"));
 }
@@ -764,6 +1069,37 @@ gnome_theme_save_clicked (GtkWidget *button,
   gnome_theme_save_show_dialog (WID ("theme_dialog"), &custom_meta_theme_info);
 }
 
+static void
+apply_font_clicked (GtkWidget *button,
+		    gpointer   data)
+{
+  GladeXML *dialog = data;
+  GConfClient *client;
+  GtkTreeSelection *selection;
+  GtkTreeModel *model;
+  GtkTreeIter iter;
+  
+  client = gconf_client_get_default ();
+  selection = gtk_tree_view_get_selection (GTK_TREE_VIEW (WID ("meta_theme_treeview")));
+  
+  if (gtk_tree_selection_get_selected (selection, &model, &iter))
+    {
+      GnomeThemeMetaInfo *meta_theme_info;
+      char *meta_theme_name;
+      
+      gtk_tree_model_get (model, &iter,
+			  META_THEME_ID_COLUMN, &meta_theme_name,
+			  -1);
+      meta_theme_info = gnome_theme_meta_info_find (meta_theme_name);
+
+      g_assert (meta_theme_info);
+      g_free (meta_theme_name);
+
+      gconf_client_set_string (client, FONT_KEY, meta_theme_info->application_font, NULL);
+    }  
+}
+	    
+	    
 static void
 setup_dialog (GladeXML *dialog)
 {
@@ -772,8 +1108,11 @@ setup_dialog (GladeXML *dialog)
   GnomeWindowManager *window_manager;
   GtkSizeGroup *size_group;
 
+  default_image = gdk_pixbuf_new_from_file(GNOMECC_DATA_DIR "/pixmaps/theme-thumbnailing.png", NULL);
+
   client = gconf_client_get_default ();
   window_manager = gnome_wm_manager_get_current (gdk_display_get_default_screen (gdk_display_get_default ()));
+  
   parent = WID ("theme_dialog");
 
   size_group = gtk_size_group_new (GTK_SIZE_GROUP_HORIZONTAL);
@@ -788,6 +1127,7 @@ setup_dialog (GladeXML *dialog)
 
   g_signal_connect (G_OBJECT (WID ("meta_theme_details_button")), "clicked", gnome_theme_details_show, NULL);
 
+  g_signal_connect (G_OBJECT (WID ("meta_theme_font1_button")), "clicked", G_CALLBACK (apply_font_clicked), dialog);
 
   setup_meta_tree_view (GTK_TREE_VIEW (WID ("meta_theme_treeview")),
 			(GCallback) meta_theme_selection_changed,
@@ -804,28 +1144,20 @@ setup_dialog (GladeXML *dialog)
 			   (GConfClientNotifyFunc) &icon_key_changed,
 			   dialog, NULL, NULL);
 
-  g_signal_connect (G_OBJECT (window_manager), "settings_changed", (GCallback) window_settings_changed, dialog);
+  gconf_client_notify_add (client,
+			   FONT_KEY,
+			   (GConfClientNotifyFunc) &font_key_changed,
+			   dialog, NULL, NULL);
+  if (window_manager)
+    g_signal_connect (G_OBJECT (window_manager),
+		      "settings_changed",
+		      (GCallback) window_settings_changed, dialog);
+
   update_themes_from_disk (dialog);
   gtk_widget_grab_focus (WID ("meta_theme_treeview"));
   gnome_theme_info_register_theme_change (theme_changed_func, dialog);
 
-  /* gtk themes */
-  widget = WID ("control_install_button");
-  g_signal_connect_swapped (G_OBJECT (widget), "clicked", G_CALLBACK (gnome_theme_installer_run), parent);
-  widget = WID ("control_manage_button");
-  g_signal_connect (G_OBJECT (widget), "clicked", G_CALLBACK (gnome_theme_manager_show_manage_themes), dialog);
 
-  /* window manager themes */
-  widget = WID ("window_install_button");
-  g_signal_connect_swapped (G_OBJECT (widget), "clicked", G_CALLBACK (gnome_theme_installer_run), parent);
-  widget = WID ("window_manage_button");
-  g_signal_connect (G_OBJECT (widget), "clicked", G_CALLBACK (gnome_theme_manager_window_show_manage_themes), dialog);
-
-  /* icon themes */
-  widget = WID ("icon_install_button");
-  g_signal_connect_swapped (G_OBJECT (widget), "clicked", G_CALLBACK (gnome_theme_installer_run), parent);
-  widget = WID ("icon_manage_button");
-  g_signal_connect (G_OBJECT (widget), "clicked", G_CALLBACK (gnome_theme_manager_show_manage_themes), dialog);
 
   widget = WID ("meta_theme_save_button");
   g_signal_connect (G_OBJECT (widget), "clicked", G_CALLBACK (gnome_theme_save_clicked), NULL);
@@ -846,7 +1178,10 @@ setup_dialog (GladeXML *dialog)
   g_signal_connect (G_OBJECT (parent), "drag-data-received",G_CALLBACK (gnome_theme_manager_drag_data_received_cb), NULL);
 
   capplet_set_icon (parent, "gnome-ccthemes.png");
+
+  update_font_button_state (dialog);
   gtk_widget_show (parent);
+
 }
 
 /* Non static functions */
@@ -861,43 +1196,68 @@ gnome_theme_manager_get_theme_dialog (void)
   return dialog;
 }
 
+gint
+gnome_theme_manager_sort_func (const gchar *a_str,
+			       const gchar *b_str,
+			       guint        a_flag,
+			       guint        b_flag)
+{
+  gint retval;
+
+  if (a_flag & (THEME_FLAG_CUSTOM | THEME_FLAG_INITIAL))
+    retval = -1;
+  else if (b_flag & (THEME_FLAG_CUSTOM | THEME_FLAG_INITIAL))
+    retval = 1;
+  else if (a_flag & THEME_FLAG_DEFAULT)
+    retval = -1;
+  else if (b_flag & THEME_FLAG_DEFAULT)
+    retval = 1;
+  else
+    {
+      retval = g_utf8_collate (a_str?a_str:"",
+			       b_str?b_str:"");
+    }
+  return retval;
+}
+
 gint
 gnome_theme_manager_tree_sort_func (GtkTreeModel *model,
 				    GtkTreeIter  *a,
 				    GtkTreeIter  *b,
 				    gpointer      user_data)
 {
-  gchar *a_str = NULL;
-  gchar *b_str = NULL;
-  guint a_default = FALSE;
-  guint b_default = FALSE;
+  GnomeThemeMetaInfo *a_meta_info;
+  GnomeThemeMetaInfo *b_meta_info;
+  gchar *a_id = NULL;
+  gchar *b_id = NULL;
+  guint a_flag = FALSE;
+  guint b_flag = FALSE;
   gint retval;
 
   gtk_tree_model_get (model, a,
-		      META_THEME_NAME_COLUMN, &a_str,
-		      META_THEME_FLAG_COLUMN, &a_default,
+		      META_THEME_ID_COLUMN, &a_id,
+		      META_THEME_FLAG_COLUMN, &a_flag,
 		      -1);
   gtk_tree_model_get (model, b,
-		      META_THEME_NAME_COLUMN, &b_str,
-		      META_THEME_FLAG_COLUMN, &b_default,
+		      META_THEME_ID_COLUMN, &b_id,
+		      META_THEME_FLAG_COLUMN, &b_flag,
 		      -1);
 
-  if (a_str == NULL) a_str = g_strdup ("");
-  if (b_str == NULL) b_str = g_strdup ("");
-
-  if (a_default & THEME_FLAG_CUSTOM)
-    retval = -1;
-  else if (b_default & THEME_FLAG_CUSTOM)
-    retval = 1;
-  else if (a_default & THEME_FLAG_DEFAULT)
-    retval = -1;
-  else if (b_default & THEME_FLAG_DEFAULT)
-    retval = 1;
+  if (a_id)
+    a_meta_info = gnome_theme_meta_info_find (a_id);
   else
-    retval = g_utf8_collate (a_str, b_str);
+    a_meta_info = &custom_meta_theme_info;
+  if (b_id)
+    b_meta_info = gnome_theme_meta_info_find (b_id);
+  else
+    b_meta_info = &custom_meta_theme_info;
 
-  g_free (a_str);
-  g_free (b_str);
+  retval = gnome_theme_manager_sort_func (a_meta_info?a_meta_info->readable_name:"",
+					  b_meta_info?b_meta_info->readable_name:"",
+					  a_flag, b_flag);
+
+  g_free (a_id);
+  g_free (b_id);
 
   return retval;
 }
@@ -925,6 +1285,29 @@ gnome_theme_manager_show_manage_themes (GtkWidget *button, gpointer data)
 	g_free (command);
 }
 
+/* Starts nautilus on the icon themes directory*/
+void
+gnome_theme_manager_icon_show_manage_themes (GtkWidget *button, gpointer data)
+{
+	gchar *path, *command;
+	GnomeVFSURI *uri;
+
+	path = g_strdup_printf ("%s/.icons", g_get_home_dir ());
+	uri = gnome_vfs_uri_new (path);
+
+	if (!gnome_vfs_uri_exists (uri)) {
+		/* Create the directory */
+		gnome_vfs_make_directory_for_uri (uri, 0775);
+	}
+	gnome_vfs_uri_unref (uri);
+
+	command = g_strdup_printf ("nautilus --no-desktop %s", path);
+	g_free (path);
+
+	g_spawn_command_line_async (command, NULL);
+	g_free (command);
+}
+
 /* Show the nautilus themes window */
 void
 gnome_theme_manager_window_show_manage_themes (GtkWidget *button, gpointer data)
@@ -986,8 +1369,16 @@ gnome_theme_manager_drag_data_received_cb (GtkWidget *widget, GdkDragContext *co
 	uris = gnome_vfs_uri_list_parse ((gchar *) selection_data->data);
 	if (uris != NULL && uris->data != NULL) {
 		GnomeVFSURI *uri = (GnomeVFSURI *) uris->data;
-		filename = gnome_vfs_unescape_string (
-			gnome_vfs_uri_get_path (uri), G_DIR_SEPARATOR_S);
+
+		if (gnome_vfs_uri_is_local (uri))
+			filename = gnome_vfs_unescape_string (
+					gnome_vfs_uri_get_path (uri),
+					G_DIR_SEPARATOR_S);
+		else
+			filename = gnome_vfs_unescape_string (
+					gnome_vfs_uri_to_string (uri, GNOME_VFS_URI_HIDE_NONE),
+					G_DIR_SEPARATOR_S);
+
 		gnome_vfs_uri_list_unref (uris);
 	}
 
@@ -1017,6 +1408,7 @@ main (int argc, char *argv[])
   activate_settings_daemon ();
 
   dialog = gnome_theme_manager_get_theme_dialog ();
+
   setup_dialog (dialog);
 
   gtk_main ();
